import glob
import os
import pandas as pd
import argparse
import numpy as np
import cv2
import time
import sys
import matplotlib.pyplot as plt

sys.path.append(os.getcwd())
from utils.utils import get_axis_aligned_bbox
from fire import Fire
from tqdm import tqdm
from tracker_gpu import SiamFCOSTracker


def main(video_dir, model_path):
    # load videos
    filenames = sorted(glob.glob(os.path.join(video_dir, "color/*.jpg")),
                       key=lambda x: int(os.path.basename(x).split('.')[0]))
    frames = [cv2.cvtColor(cv2.imread(filename), cv2.COLOR_BGR2RGB) for filename in filenames]

    gt_bboxes = np.loadtxt(os.path.join(video_dir, "groundtruth.txt"), delimiter=',', dtype='f8')

    title = video_dir.split('/')[-1]
    if not os.path.isdir('results/' + title):
        os.makedirs('results/' + title)
    # starting tracking
    tracker = SiamFCOSTracker(model_path)
    for idx, frame in tqdm(enumerate(frames), total=len(frames)):
        # name = 'results/' +title + '/' + str(idx)
        if idx == 0:
            gt_bbox = gt_bboxes[0]
            [cx, cy, w, h] = get_axis_aligned_bbox(gt_bbox)
            bbox = [cx - w/2 + 0.5, cy - h/2 + 0.5, w, h]
            z = tracker.init(frame, bbox)
            bbox = (bbox[0], bbox[1],
                    bbox[0] + bbox[2] - 1, bbox[1] + bbox[3] - 1)
        else:
            bbox, score_pred, _, fore_pixels5, fore_pixels9, p_95 = tracker.update(frame)  # bbox为中心和宽高
            bbox = (bbox[0] - bbox[2] / 2 + 0.5, bbox[1] - bbox[3] / 2 + 0.5,
                    bbox[0] + bbox[2] / 2 - 0.5, bbox[1] + bbox[3] / 2 - 0.5)
            # cv2.imwrite('results/' + title + '/{:0>4d}_instance.jpg'.format(idx), see)
            #             cv2.resize(target, (127, 127), interpolation=cv2.INTER_LINEAR))
        # bbox xmin ymin xmax ymax

        frame = cv2.rectangle(frame,
                              (int(bbox[0]), int(bbox[1])),
                              (int(bbox[2]), int(bbox[3])),
                              (0, 255, 0),
                              2)
        gt_bbox = gt_bboxes[idx]
        [cx, cy, w, h] = get_axis_aligned_bbox(gt_bbox)
        gt_bbox = [cx - w / 2 + 0.5, cy - h / 2 + 0.5, w, h]
        gt_bbox = (gt_bbox[0], gt_bbox[1],
                   gt_bbox[0] + gt_bbox[2] - 1, gt_bbox[1] + gt_bbox[3] - 1)
        if (np.array(gt_bbox) != np.array(gt_bbox)).any():  # 不为nan
            pass
        else:
            frame = cv2.rectangle(frame,
                                  (int(gt_bbox[0]), int(gt_bbox[1])),  # 0-index
                                  (int(gt_bbox[2]), int(gt_bbox[3])),
                                  (255, 0, 0),
                                  1)
        if len(frame.shape) == 3:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        frame = cv2.putText(frame, str(idx), (5, 20), cv2.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 255, 0), 1)
        if idx != 0:
            for point in list(fore_pixels5):
                cv2.circle(frame, (int(point[0][0]), int(point[0][1])), 1, (0, 0, 255), 1)
            for point in list(fore_pixels9):
                cv2.circle(frame, (int(point[0][0]), int(point[0][1])), 1, (255, 0, 0), 1)
            if (int(bbox[1] > 10)):
                cv2.putText(frame, str(p_95), (int(bbox[0]), int(bbox[1] - 6)), cv2.FONT_HERSHEY_COMPLEX_SMALL,1 ,
                            (255, 255, 0), 1)
            else:
                cv2.putText(frame, str(p_95), (int(bbox[0]), int(bbox[1] + 15)), cv2.FONT_HERSHEY_COMPLEX_SMALL, 1,
                            (255, 255, 0), 1)

        cv2.imwrite('results/' + title + '/{:0>4d}.jpg'.format(idx), frame)


if __name__ == "__main__":
    Fire(main)
