import lmdb
import cv2
import numpy as np
import os
import hashlib
import functools

from glob import glob
from fire import Fire
from tqdm import tqdm
from multiprocessing import Pool
from IPython import embed
import multiprocessing as mp
import xml.etree.ElementTree as ET
import pickle
import shutil

# coco_class = ['person', 'bicycle', 'car', 'motorcycle','airplane',
#                 'bus',  'train',  'truck',  'boat',  'traffic light',
#                 'fire hydrant',  'stop sign',  'parking meter', 'bench',
#                 'bird', 'cat',  'dog',  'horse',  'sheep',  'cow',
#                 'elephant',  'bear',  'zebra',  'giraffe',  'backpack',
#                 'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee',
#                 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
#                 'baseball glove', 'skateboard', 'surfboard', 'tennis racket',
#                 'bottle', 'wine glass', 'cup', 'fork', 'knife', 'spoon',
#                 'bowl', 'banana', 'apple', 'sandwich', 'orange',
#                 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut',
#                 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table',
#                 'toilet', 'tv', 'laptop', 'mouse', 'remote', 'keyboard',
#                 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
#                 'refrigerator', 'book', 'clock', 'vase', 'scissors',
#                 'teddy bear', 'hair drier', 'toothbrush']
coco_class = ['person', 'bicycle', 'car', 'motorcycle','airplane',
                'bus', 'truck', 'boat', 'parking meter', 'bench',
                'bird', 'cat',  'dog',  'horse',  'sheep',  'cow',
                'elephant',  'bear',  'zebra',  'giraffe',  'backpack',
                'umbrella', 'handbag', 'suitcase', 'frisbee',
                'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                'baseball glove', 'skateboard', 'surfboard', 'tennis racket',
                'bottle', 'wine glass', 'fork', 'book', 'teddy bear', 'hair drier']

def darker(image,percetage=0.9):
    image_copy = image.copy()
    w = image.shape[1]
    h = image.shape[0]
    #get darker
    for xi in range(0,w):
        for xj in range(0,h):
            image_copy[xj,xi,0] = int(image[xj,xi,0]*percetage)
            image_copy[xj,xi,1] = int(image[xj,xi,1]*percetage)
            image_copy[xj,xi,2] = int(image[xj,xi,2]*percetage)
    return image_copy
def brighter(image, percetage=1.5):
    image_copy = image.copy()
    w = image.shape[1]
    h = image.shape[0]
    #get brighter
    for xi in range(0,w):
        for xj in range(0,h):
            image_copy[xj,xi,0] = np.clip(int(image[xj,xi,0]*percetage),a_max=255,a_min=0)
            image_copy[xj,xi,1] = np.clip(int(image[xj,xi,1]*percetage),a_max=255,a_min=0)
            image_copy[xj,xi,2] = np.clip(int(image[xj,xi,2]*percetage),a_max=255,a_min=0)
    return image_copy
def rotate(image, angle=10, scale=0.9):
    w = image.shape[1]
    h = image.shape[0]
    #rotate matrix
    M = cv2.getRotationMatrix2D((w/2,h/2), angle, scale)
    #rotate
    image = cv2.warpAffine(image,M,(w,h))
    return image


def worker(video_name):
    image_names = glob(video_name + '/*')
    kv = {}
    for image_name in image_names:
        img = cv2.imread(image_name)
        # 数据增强
        if np.random.random() > 0.5:
            img = cv2.flip(img,1)
        if np.random.random() > 0.5:
            a = np.random.uniform(0.5, 1.5)
            b = np.random.uniform(-10, 10)
            img = np.uint8(np.clip((a * img + b), 0, 255))
        if np.random.random() > 0.5:
            img = rotate(img)
        if np.random.random() > 0.65:
            if np.random.random() > 0.4:
                kernel_size = (3,3)
            else:
                kernel_size = (5, 5)
            sigma = 1
            img = cv2.GaussianBlur(img, kernel_size, sigma)


        _, img_encode = cv2.imencode('.jpg', img)
        img_encode = img_encode.tobytes()
        kv[hashlib.md5(image_name.encode()).digest()] = img_encode
    return kv

def pkl():
    meta_data_path1 = '/media/lyc/data/Save_data1/meta_data.pkl'
    meta_data_path2 = '/media/lyc/data/Save_data/meta_data.pkl'
    meta_data = []
    meta_data1 = pickle.load(open(meta_data_path1, 'rb'))
    meta_data2 = pickle.load(open(meta_data_path2, 'rb'))
    meta_data = meta_data1 + meta_data2[:-1]

    return meta_data


def create_lmdb(data_dir, output_dir, num_threads=mp.cpu_count()):
    meta_data_path1 = '/media/lyc/data/Save_data1/meta_data.pkl'
    meta_data_path2 = '/media/lyc/data/meta_data.pkl'
    meta_data1 = pickle.load(open(meta_data_path1, 'rb'))
    meta_data2 = pickle.load(open(meta_data_path2, 'rb'))
    meta_data = meta_data1[:28888] + meta_data2[:-1]

    video_names = glob(data_dir + '/*')
    video_names = [x for x in video_names if 'meta_data.pkl' not in x]

    video_name_list = []
    # meta_data_path = os.path.join(data_dir, 'meta_data.pkl')
    # meta_data = pickle.load(open(meta_data_path, 'rb'))
    meta_copy = meta_data.copy()
    trajs = {}
    flag = 0
    for data in meta_copy:
        a = data[0]
        if 'COCO' in data[0]:
           if data[1] == {}:
               meta_data.remove(data)
           else:
               idx = video_names.index(os.path.join(data_dir, data[0]))
               video_name_list.append(video_names[idx])
        else:
            idx = video_names.index(os.path.join(data_dir, data[0]))
            video_name_list.append(video_names[idx])

    pickle.dump(meta_data, open(os.path.join('/media/lyc/data/Save_data/', "meta_data.pkl"), 'wb'))
    # video_names = [x for x in video_names if os.path.isdir(x)]
    db = lmdb.open(output_dir, map_size=int(200e9))
    CN = mp.cpu_count()

    with Pool(processes=CN) as pool:

        for ret in tqdm(pool.imap_unordered(functools.partial(worker), video_name_list), total=len(video_name_list)):
            with db.begin(write=True) as txn:
                for k, v in ret.items():
                    txn.put(k, v)


if __name__ == '__main__':
    Fire(create_lmdb)
    # pkl()