import torch
import numpy as np
import cv2

def get_center(x):
    return (x - 1.) / 2.


def xyxy2cxcywh(bbox):
    return get_center(bbox[0] + bbox[2]), \
           get_center(bbox[1] + bbox[3]), \
           (bbox[2] - bbox[0]), \
           (bbox[3] - bbox[1])


def cxcywh2xyxy(bboxes):
    if len(np.array(bboxes).shape) == 1:
        bboxes = np.array(bboxes)[None, :]
    else:
        bboxes = np.array(bboxes)
    x1 = bboxes[:, 0:1] + 1 / 2 - bboxes[:, 2:3] / 2
    x2 = x1 + bboxes[:, 2:3] - 1
    y1 = bboxes[:, 1:2] + 1 / 2 - bboxes[:, 3:4] / 2
    y2 = y1 + bboxes[:, 3:4] - 1
    return np.concatenate([x1, y1, x2, y2], 1)


def nms(bboxes, scores, num, threshold=0.7):
    sort_index = np.argsort(scores)[::-1]
    sort_boxes = bboxes[sort_index]
    selected_bbox = [sort_boxes[0]]
    selected_index = [sort_index[0]]
    for i, bbox in enumerate(sort_boxes):
        iou = compute_iou(selected_bbox, bbox)
        # print(iou, bbox, selected_bbox)
        if np.max(iou) < threshold:
            selected_bbox.append(bbox)
            selected_index.append(sort_index[i])
            if len(selected_bbox) >= num:
                break
    return selected_index


def nms_worker(x, threshold=0.7):
    bboxes, scores, num = x
    if len(bboxes) == 0:
        selected_index = [0]
    else:
        sort_index = np.argsort(scores)[::-1]
        sort_boxes = bboxes[sort_index]
        selected_bbox = [sort_boxes[0]]
        selected_index = [sort_index[0]]
        for i, bbox in enumerate(sort_boxes):
            iou = compute_iou(selected_bbox, bbox)
            # print(iou, bbox, selected_bbox)
            if np.max(iou) < threshold:
                selected_bbox.append(bbox)
                selected_index.append(sort_index[i])
                if len(selected_bbox) >= num:
                    break
    return selected_index


def round_up(value):
    # 替换内置round函数,实现保留2位小数的精确四舍五入
    return round(value + 1e-6 + 1000) - 1000


def crop_and_pad(img, cx, cy, model_sz, original_sz, img_mean=None):
    # model_sz = 127,  original_sz = bbox扩大大约两倍
    # resize操作 original_sz 到 model_sz

    # 对于instance img， model_sz=original_sz=271
    im_h, im_w, _ = img.shape

    xmin = cx - (original_sz - 1) / 2
    xmax = xmin + original_sz - 1
    ymin = cy - (original_sz - 1) / 2
    ymax = ymin + original_sz - 1

    left = int(round_up(max(0., -xmin)))
    top = int(round_up(max(0., -ymin)))
    right = int(round_up(max(0., xmax - im_w + 1)))
    bottom = int(round_up(max(0., ymax - im_h + 1)))

    xmin = int(round_up(xmin + left)) # 0 或 xmin
    xmax = int(round_up(xmax + left))
    ymin = int(round_up(ymin + top))
    ymax = int(round_up(ymax + top))
    r, c, k = img.shape
    if any([top, bottom, left, right]):
        te_im = np.zeros((r + top + bottom, c + left + right, k), np.uint8)  # 0 is better than 1 initialization
        te_im[top:top + r, left:left + c, :] = img
        if top:
            te_im[0:top, left:left + c, :] = img_mean
        if bottom:
            te_im[r + top:, left:left + c, :] = img_mean
        if left:
            te_im[:, 0:left, :] = img_mean
        if right:
            te_im[:, c + left:, :] = img_mean
        im_patch_original = te_im[int(ymin):int(ymax + 1), int(xmin):int(xmax + 1), :]
    else:
        im_patch_original = img[int(ymin):int(ymax + 1), int(xmin):int(xmax + 1), :]
    if not np.array_equal(model_sz, original_sz):
        im_patch = cv2.resize(im_patch_original, (model_sz, model_sz))  # zzp: use cv to get a better speed
    else:
        im_patch = im_patch_original
    scale = model_sz / im_patch_original.shape[0]
    return im_patch, scale

def crop_and_pad1(img, cx, cy, model_sz, original_wz, original_hz, img_mean=None):
    # model_sz = 127,  original_sz = bbox扩大大约两倍
    # resize操作 original_sz 到 model_sz

    # 对于instance img， model_sz=original_sz=271
    im_h, im_w, _ = img.shape

    xmin = cx - (original_wz - 1) / 2
    xmax = xmin + original_wz - 1
    ymin = cy - (original_hz - 1) / 2
    ymax = ymin + original_hz - 1

    left = int(round_up(max(0., -xmin)))
    top = int(round_up(max(0., -ymin)))
    right = int(round_up(max(0., xmax - im_w + 1)))
    bottom = int(round_up(max(0., ymax - im_h + 1)))

    xmin = int(round_up(xmin + left)) # 0 或 xmin
    xmax = int(round_up(xmax + left))
    ymin = int(round_up(ymin + top))
    ymax = int(round_up(ymax + top))
    r, c, k = img.shape
    if any([top, bottom, left, right]):
        te_im = np.zeros((r + top + bottom, c + left + right, k), np.uint8)  # 0 is better than 1 initialization
        te_im[top:top + r, left:left + c, :] = img
        if top:
            te_im[0:top, left:left + c, :] = img_mean
        if bottom:
            te_im[r + top:, left:left + c, :] = img_mean
        if left:
            te_im[:, 0:left, :] = img_mean
        if right:
            te_im[:, c + left:, :] = img_mean
        im_patch_original = te_im[int(ymin):int(ymax + 1), int(xmin):int(xmax + 1), :]
    else:
        im_patch_original = img[int(ymin):int(ymax + 1), int(xmin):int(xmax + 1), :]
    # if not np.array_equal(model_sz, original_wz):
    im_patch = cv2.resize(im_patch_original, (model_sz, model_sz))  # zzp: use cv to get a better speed
    # else:
    #     im_patch = im_patch_original
    scale_w = model_sz / im_patch_original.shape[0]
    scale_h = model_sz / im_patch_original.shape[1]
    return im_patch, scale_w, scale_h

def get_exemplar_image1(img, bbox, exemplar_size):
    size_z = exemplar_size
    cx, cy, w, h = bbox
    # wc_z = w + context_amount * (w + h)
    # hc_z = h + context_amount * (w + h)
    # s_z = np.sqrt(wc_z * hc_z)
    # scale_z = size_z / s_z  # 放大的倍数
    x_left_t = int(round_up(cx - (w / 2)))  # 列
    x_right_t = int(round_up(cx + (w / 2)))  # 列
    y_up_t = int(round_up(cy - (h / 2)))  # 行
    y_down_t = int(round_up(cy + (h / 2)))  # 行
    gt_img = img[y_up_t:y_down_t, x_left_t:x_right_t]

    #下面语句为将gt_img resize成为127*127大小的图
    # gt_img -> exemplar_img
    dim = (size_z,size_z)
    exemplar_img = cv2.resize(gt_img,dim)

    scale_w = w / size_z
    scale_h = h / size_z
    # exemplar_img, _ = crop_and_pad(img, cx, cy, size_z, s_z, img_mean)
    return exemplar_img, scale_w, scale_h


def get_instance_image1(img, bbox, instance_size, scale_w, scale_h, img_mean=None):
    # context_amount = 0.5
    cx, cy, w, h = bbox  # float type
    # wc_z = w + context_amount * (w + h)
    # hc_z = h + context_amount * (w + h)
    # s_z = np.sqrt(wc_z * hc_z)  # the width of the crop box
    # scale_z = size_z / s_z
    #
    # s_x = s_z * size_x / size_z # sx 为sz 的两倍，为应该切的区域大小
    # instance_img, scale_x = crop_and_pad(img, cx, cy, size_x, s_x, img_mean)  # 与z形变一致
    # w_x = w * scale_x
    # h_x = h * scale_x
    # point_1 = (size_x + 1) / 2 - w_x / 2, (size_x + 1) / 2 - h_x / 2
    # point_2 = (size_x + 1) / 2 + w_x / 2, (size_x + 1) / 2 + h_x / 2
    # frame = cv2.rectangle(instance_img, (int(point_1[0]),int(point_1[1])), (int(point_2[0]),int(point_2[1])), (0, 255, 0), 2)
    # cv2.imwrite('1.jpg', frame)
    size_x = instance_size
    w_x = scale_w * size_x
    h_x = scale_h * size_x
    # x_left_i = int(round_up(cx - (w_x / 2)))  # 列
    # x_right_i = int(round_up(cx + (w_x / 2)))  # 列
    # y_up_i = int(round_up(cy - (h_x / 2)))  # 行
    # y_down_i = int(round_up(cy + (h_x / 2)))  #
    instance_img, scale_w, scale_h = crop_and_pad1(img, cx, cy, size_x, w_x, h_x, img_mean)
    # instance_img = img[y_up_i:y_down_i, x_left_i:x_right_i]
    return instance_img, scale_w, scale_h




def get_exemplar_image(img, bbox, size_z, context_amount, img_mean=None):
    # size_z = 127
    cx, cy, w, h = bbox
    wc_z = w + context_amount * (w + h)
    hc_z = h + context_amount * (w + h)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = size_z / s_z  # 放大的倍数
    exemplar_img, _ = crop_and_pad(img, cx, cy, size_z, s_z, img_mean)
    return exemplar_img, scale_z, s_z


def get_instance_image(img, bbox, size_z, size_x, context_amount, img_mean=None):
    # context_amount = 0.5
    cx, cy, w, h = bbox  # float type
    wc_z = w + context_amount * (w + h)
    hc_z = h + context_amount * (w + h)
    s_z = np.sqrt(wc_z * hc_z)  # the width of the crop box
    scale_z = size_z / s_z

    s_x = s_z * size_x / size_z # sx 为sz 的两倍，为应该切的区域大小
    instance_img, scale_x = crop_and_pad(img, cx, cy, size_x, s_x, img_mean)  # 与z形变一致
    w_x = w * scale_x
    h_x = h * scale_x
    # point_1 = (size_x + 1) / 2 - w_x / 2, (size_x + 1) / 2 - h_x / 2
    # point_2 = (size_x + 1) / 2 + w_x / 2, (size_x + 1) / 2 + h_x / 2
    # frame = cv2.rectangle(instance_img, (int(point_1[0]),int(point_1[1])), (int(point_2[0]),int(point_2[1])), (0, 255, 0), 2)
    # cv2.imwrite('1.jpg', frame)
    return instance_img, w_x, h_x, scale_x

def get_instance_image_detect(img, bbox, size_z, size_x1, size_x, context_amount, img_mean=None):
    # context_amount = 0.5
    cx, cy, w, h = bbox  # float type
    wc_z = w + context_amount * (w + h)
    hc_z = h + context_amount * (w + h)
    s_z = np.sqrt(wc_z * hc_z)  # the width of the crop box
    scale_z = size_z / s_z

    s_x = s_z * size_x / size_z # sx 为sz 的两倍，为应该切的区域大小
    instance_img, scale_x = crop_and_pad(img, cx, cy, size_x1, s_x, img_mean)  # 与z形变一致
    w_x = w * scale_x
    h_x = h * scale_x
    # point_1 = (size_x + 1) / 2 - w_x / 2, (size_x + 1) / 2 - h_x / 2
    # point_2 = (size_x + 1) / 2 + w_x / 2, (size_x + 1) / 2 + h_x / 2
    # frame = cv2.rectangle(instance_img, (int(point_1[0]),int(point_1[1])), (int(point_2[0]),int(point_2[1])), (0, 255, 0), 2)
    # cv2.imwrite('1.jpg', frame)
    return instance_img, w_x, h_x, scale_x

def box_transform(anchors, gt_box):
    anchor_xctr = anchors[:, :1]
    anchor_yctr = anchors[:, 1:2]
    gt_cx, gt_cy, gt_w, gt_h = gt_box

    target_l = (anchor_xctr - (gt_cx - gt_w/2 + 0.5)) #/ img_sz
    target_t = (anchor_yctr - (gt_cy - gt_h/2 + 0.5)) #/ img_sz
    target_r = ((gt_cx + gt_w/2 - 0.5) - anchor_xctr) #/ img_sz
    target_b = ((gt_cy + gt_h/2 - 0.5) - anchor_yctr) #/ img_sz
    regression_target = np.hstack((target_l, target_t, target_r, target_b))

    lr = np.hstack((target_l, target_r))
    tb = np.hstack((target_t, target_b))
    c = (np.min(lr,axis=1)/ np.max(lr,axis=1))*(np.min(tb,axis=1)/ np.max(tb,axis=1))
    c[(target_l[:,0]<0)]=0
    c[(target_r[:,0]<0)]=0
    c[(target_t[:,0]<0)]=0
    c[(target_b[:,0]<0)]=0
    # target = c.reshape(23, 23, 1)
    # centerness = np.hstack(c)

    return regression_target, c


def box_transform_inv(anchors, offset, img_sz=271):
    anchor_xctr = anchors[:, :1]
    anchor_yctr = anchors[:, 1:2]

    x = anchor_xctr #+ img_sz // 2
    y = anchor_yctr #+ img_sz // 2

    offset_l, offset_t, offset_r, offset_b = offset[:, :1], offset[:, 1:2], offset[:, 2:3], offset[:, 3:]

    # box_cx = x
    # box_cy = y
    # box_l = offset_l#*img_sz
    # box_r = offset_r#*img_sz
    # box_t = offset_t#*img_sz
    # box_b = offset_b#*img_sz
    box_cx = (x-offset_l+x+offset_r)/2
    box_cy = (y-offset_t+y+offset_b)/2
    box_w = (offset_l + offset_r)+1 # * img_sz
    box_h = (offset_t + offset_b)+1 # * img_sz
    box = np.hstack([box_cx, box_cy, box_w, box_h])
    return box


def get_topk_box(cls_score, pred_regression, anchors, img_sz, topk=10):
    # anchors xc,yc,w,h
    regress_offset = pred_regression.cpu().detach().numpy()

    scores, index = torch.topk(cls_score, topk, )
    index = index.view(-1).cpu().detach().numpy()

    topk_offset = regress_offset[index, :]
    anchors = anchors[index, :]
    pred_box = box_transform_inv(anchors, img_sz, topk_offset)
    return pred_box


def compute_iou(anchors, box):
    if np.array(anchors).ndim == 1:
        anchors = np.array(anchors)[None, :]
    else:
        anchors = np.array(anchors)
    if np.array(box).ndim == 1:
        box = np.array(box)[None, :]
    else:
        box = np.array(box)
    gt_box = np.tile(box.reshape(1, -1), (anchors.shape[0], 1))

    anchor_x1 = anchors[:, :1] - anchors[:, 2:3] / 2 + 0.5
    anchor_x2 = anchors[:, :1] + anchors[:, 2:3] / 2 - 0.5
    anchor_y1 = anchors[:, 1:2] - anchors[:, 3:] / 2 + 0.5
    anchor_y2 = anchors[:, 1:2] + anchors[:, 3:] / 2 - 0.5

    gt_x1 = gt_box[:, :1] - gt_box[:, 2:3] / 2 + 0.5
    gt_x2 = gt_box[:, :1] + gt_box[:, 2:3] / 2 - 0.5
    gt_y1 = gt_box[:, 1:2] - gt_box[:, 3:] / 2 + 0.5
    gt_y2 = gt_box[:, 1:2] + gt_box[:, 3:] / 2 - 0.5

    xx1 = np.max([anchor_x1, gt_x1], axis=0)
    xx2 = np.min([anchor_x2, gt_x2], axis=0)
    yy1 = np.max([anchor_y1, gt_y1], axis=0)
    yy2 = np.min([anchor_y2, gt_y2], axis=0)

    inter_area = np.max([xx2 - xx1, np.zeros(xx1.shape)], axis=0) * np.max([yy2 - yy1, np.zeros(xx1.shape)],
                                                                           axis=0)
    area_anchor = (anchor_x2 - anchor_x1) * (anchor_y2 - anchor_y1)
    area_gt = (gt_x2 - gt_x1) * (gt_y2 - gt_y1)
    iou = inter_area / (area_anchor + area_gt - inter_area + 1e-6)
    return iou


def get_pyramid_instance_image(img, center, size_x, size_x_scales, img_mean=None):
    if img_mean is None:
        img_mean = tuple(map(int, img.mean(axis=(0, 1))))
    pyramid = [crop_and_pad(img, center[0], center[1], size_x, size_x_scale, img_mean)
               for size_x_scale in size_x_scales]
    return pyramid

def get_axis_aligned_bbox(region):
    try:
        region = np.array([region[0][0][0], region[0][0][1], region[0][1][0], region[0][1][1],
                           region[0][2][0], region[0][2][1], region[0][3][0], region[0][3][1]])
    except:
        region = np.array(region)
    cx = np.mean(region[0::2])
    cy = np.mean(region[1::2])
    x1 = min(region[0::2])
    x2 = max(region[0::2])
    y1 = min(region[1::2])
    y2 = max(region[1::2])
    A1 = np.linalg.norm(region[0:2] - region[2:4]) * np.linalg.norm(region[2:4] - region[4:6])
    A2 = (x2 - x1) * (y2 - y1)
    s = np.sqrt(A1 / A2)
    w = s * (x2 - x1) + 1
    h = s * (y2 - y1) + 1
    return cx, cy, w, h

def add_box_img(img, boxes, color=(0, 255, 0)):
    # boxes (x,y,w,h)
    if boxes.ndim == 1:
        boxes = boxes[None, :]
    img = img.copy()
    img_ctx = (img.shape[1] - 1) / 2
    img_cty = (img.shape[0] - 1) / 2
    for box in boxes:
        point_1 = [img_ctx - box[2] / 2 + box[0] + 0.5, img_cty - box[3] / 2 + box[1] + 0.5]
        point_2 = [img_ctx + box[2] / 2 + box[0] - 0.5, img_cty + box[3] / 2 + box[1] - 0.5]
        point_1[0] = np.clip(point_1[0], 0, img.shape[1])
        point_2[0] = np.clip(point_2[0], 0, img.shape[1])
        point_1[1] = np.clip(point_1[1], 0, img.shape[0])
        point_2[1] = np.clip(point_2[1], 0, img.shape[0])
        img = cv2.rectangle(img, (int(point_1[0]), int(point_1[1])), (int(point_2[0]), int(point_2[1])),
                            color, 2)
    return img


def adjust_learning_rate(optimizer, decay=0.1):
    """Sets the learning rate to the initial LR decayed by 0.5 every 20 epochs"""
    for param_group in optimizer.param_groups:
        param_group['lr'] = decay * param_group['lr']
def add_box_img_left_top(img, boxes, color=(0, 255, 0)):
    # boxes (x,y,w,h)
    if boxes.ndim == 1:
        boxes = boxes[None, :]
    img = img.copy()
    for i, box in enumerate(boxes):
        point_1 = [- box[2] / 2 + box[0] + 0.5, - box[3] / 2 + box[1] + 0.5]
        point_2 = [+ box[2] / 2 + box[0] - 0.5, + box[3] / 2 + box[1] - 0.5]
        point_1[0] = np.clip(point_1[0], 0, img.shape[1])
        point_2[0] = np.clip(point_2[0], 0, img.shape[1])
        point_1[1] = np.clip(point_1[1], 0, img.shape[0])
        point_2[1] = np.clip(point_2[1], 0, img.shape[0])
        img = cv2.rectangle(img, (int(point_1[0]), int(point_1[1])), (int(point_2[0]), int(point_2[1])),
                            color, 2)
    return img

def cxy_wh_2_rect(pos, sz):
    return np.array([pos[0]-sz[0]/2, pos[1]-sz[1]/2, sz[0], sz[1]])  # 0-index